from dataclasses import dataclass, field
from enum import StrEnum, auto

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


@dataclass
class Result:
    status_code: int
    message: str = ""
    data: dict = field(default_factory=dict)


class GovinfoModel(BaseModel):
    # TODO: move shared attributes into this model,
    # e.g., dateIssued, packageId, collectionCode, collectionName, docClass, category?
    # download?, last_modified, title?
    model_config = ConfigDict(alias_generator=to_camel)


class Branch(StrEnum):
    LEGISLATIVE = auto()
    EXECUTIVE = auto()
    JUDICIAL = auto()


# NOTE: response objects from `related` endpoints and `packages` summary endpoints are similarly vague



from govinfo.config import RequestArgs
from govinfo.exceptions import GovinfoException
from govinfo.models import GovinfoModel
from govinfo.packages import PackageInfo


class SummaryItem(GovinfoModel):
    collection_code: str
    collection_name: str
    package_count: int
    granule_count: int | None


class CollectionSummary(GovinfoModel):
    collections: list[SummaryItem]


class CollectionContainer(GovinfoModel):
    count: int
    message: str | None
    next_page: str | None
    previous_page: str | None
    packages: list[PackageInfo]


class CollectionsMixin:
    def _build_collections_request(
        self,
        collection: str = None,
        start_date: str = None,
        end_date: str = None,
        **kwargs,
    ) -> RequestArgs:
        endpoint_parts = ["collections", collection, start_date, end_date]
        path = "/".join(part for part in endpoint_parts if part is not None)
        params = self._set_params(**kwargs)
        return (path, params)

    def collections(
        self,
        collection: str = None,
        start_date: str = None,
        end_date: str = None,
        **kwargs,
    ):
        """Call the collections endpoint of the GovInfo API."""
        args = self._build_collections_request(
            collection, start_date, end_date, **kwargs
        )

        try:
            result = self._get(args)
        except GovinfoException as e:
            raise e

        if collection is None:
            validated = CollectionSummary(**result.data)
        else:
            validated = CollectionContainer(**result.data)
        # TODO: dump with(out) alias?
        return validated.model_dump()



from govinfo import Govinfo
from govinfo.config import OFFSET_DEFAULT, PAGE_DEFAULT


def test_govinfo_default_api_key():
    govinfo = Govinfo()
    assert govinfo._api_key == "DEMO_KEY"


def test_govinfo_user_supplied_api_key():
    govinfo = Govinfo(api_key="dummy key")
    assert govinfo._api_key == "dummy key"


def test_govinfo_repr():
    govinfo = Govinfo()
    assert str(govinfo) == "Govinfo(url='https://api.govinfo.gov', api_key='DEMO_KEY')"
    govinfo = Govinfo(api_key="dummy key")
    assert (
        str(govinfo)
        == "Govinfo(url='https://api.govinfo.gov', api_key='user supplied')"
    )


def test_build_default_collections_request():
    govinfo = Govinfo()
    path, params = govinfo._build_collections_request()
    assert path == "collections"
    assert params == {"offsetMark": OFFSET_DEFAULT, "pageSize": PAGE_DEFAULT}


def test_build_collections_request_with_args():
    govinfo = Govinfo()
    path, params = govinfo._build_collections_request(
        "bills", "2025-06-16T00:00:00Z", page_size=10, offset_mark="something"
    )
    assert path == "collections/bills/2025-06-16T00:00:00Z"
    assert params == {"offsetMark": "something", "pageSize": 10}


def test_build_default_packages_granules_request():
    govinfo = Govinfo()
    path, params = govinfo._build_granules_request("CREC-2018-01-04")
    assert path == "packages/CREC-2018-01-04/granules"
    assert params == {"offsetMark": OFFSET_DEFAULT, "pageSize": PAGE_DEFAULT}


def test_build__packages_granules_request_with_args():
    govinfo = Govinfo()
    path, params = govinfo._build_granules_request(
        "CREC-2018-01-04", granule_class="something", md5="something"
    )
    assert path == "packages/CREC-2018-01-04/granules"
    assert params == {
        "offsetMark": OFFSET_DEFAULT,
        "pageSize": PAGE_DEFAULT,
        "granuleClass": "something",
        "md5": "something",
    }



from json import JSONDecodeError

import httpx

from govinfo.collections import CollectionsMixin
from govinfo.config import BASE_URL, OFFSET_DEFAULT, PAGE_DEFAULT, RequestArgs
from govinfo.exceptions import GovinfoException
from govinfo.models import Result
from govinfo.packages import PackagesMixin


class Govinfo(CollectionsMixin, PackagesMixin):
    """Wrapper class for the GovInfo API.

    Users can supply an API key or use the default value, DEMO_KEY"""

    def __init__(self, api_key: str = "DEMO_KEY"):
        self._url = f"{BASE_URL}"
        self._api_key = api_key

    def _get(self, args: RequestArgs) -> Result:
        headers = {"x-api-key": self._api_key}
        path, params = args
        with httpx.Client(base_url=self._url, headers=headers, params=params) as client:
            response = client.get(path)
            try:
                data = response.json()
            except (ValueError, JSONDecodeError) as e:
                raise GovinfoException("Bad JSON in response") from e
            is_success = 299 >= response.status_code >= 200
            if is_success:
                return Result(
                    response.status_code, message=response.reason_phrase, data=data
                )
            raise GovinfoException(f"{response.status_code}: {response.reason_phrase}")

    def __repr__(self) -> str:
        api_key = "user supplied" if self._is_api_key_set() else self._api_key
        return f"Govinfo(url={self._url!r}, api_key={api_key!r})"

    def _is_api_key_set(self) -> bool:
        return self._api_key != "DEMO_KEY"

    def _set_params(self, *args, **kwargs) -> dict[str, str]:
        default_params = {"offsetMark": OFFSET_DEFAULT, "pageSize": PAGE_DEFAULT}
        params = (
            default_params
            if not kwargs
            else default_params
            | {
                key.split("_")[0]
                + "".join(word.capitalize() for word in key.split("_")[1:]): value
                for key, value in kwargs.items()
            }
        )
        return params



from datetime import date, datetime

from pydantic import ConfigDict, Field

from govinfo.config import RequestArgs
from govinfo.exceptions import GovinfoException
from govinfo.models import Branch, GovinfoModel


class PackageInfo(GovinfoModel):
    package_id: str
    last_modified: datetime
    package_link: str
    doc_class: str
    title: str
    congress: int
    date_issued: date


class GranuleMetadata(GovinfoModel):
    title: str
    granule_id: str
    granule_link: str
    granule_class: str
    md5: str = Field(default=None)


class GranuleContainer(GovinfoModel):
    count: int
    offset: int | None
    page_size: int
    next_page: str | None
    previous_page: str | None
    granules: list[GranuleMetadata]
    message: str = Field(default=None)


class PackageSummary(GovinfoModel):
    # "allow" since there are so many variations on what is returned
    model_config = ConfigDict(extra="allow")
    category: str
    date_issued: date
    collection_code: str
    collection_name: str
    doc_class: str
    publisher: str
    last_modfied: datetime
    branch: Branch
    # TODO: specify download model
    download: dict
    # TODO: specify other_identifier model
    other_identifier: dict


# TODO: create models that inherit from PackageSummary for specific packages types
# start with BILLS, PLAW, CREC, CRPT, CPRT


class PackagesMixin:
    def _build_granules_request(
        self,
        package_id: str,
        **kwargs,
    ) -> RequestArgs:
        path = f"packages/{package_id}/granules"
        params = self._set_params(**kwargs)
        return (path, params)

    def granules(self, package_id: str, **kwargs):
        """Call the packages/{package_id}/granules endpoint of the GovInfo API."""
        args = self._build_granules_request(package_id, **kwargs)

        try:
            result = self._get(args)
        except GovinfoException as e:
            raise e

        validated = GranuleContainer(**result.data)
        return validated.model_dump()
